__author__ = "Vanessa Sochat, Alec Scott"
__copyright__ = "Copyright 2021-2022, Vanessa Sochat and Alec Scott"
__license__ = "Apache-2.0"


def parse_extra(extras):
    """
    Given a listing of extra arguments, parse into lookup dict.
    """
    values = {}
    if not extras:
        return values

    while extras:
        key = extras.pop(0)

        # A boolean or a key/value pair
        if key.startswith("--") and extras:
            key = key.replace("--", "").replace("-", "_")
            value = extras.pop(0)

            # key value pair
            if not value.startswith("--"):
                values[key] = value

            # Two booleans
            else:
                values[key] = True
                key = value.replace("--", "").replace("-", "_")
                values[key] = True

        # Just a boolean flag
        elif key.startswith("--") and not extras:
            key = key.replace("--", "").replace("-", "_")
            extras[key] = True
    return values



__author__ = "Vanessa Sochat, Alec Scott"
__copyright__ = "Copyright 2021-2022, Vanessa Sochat and Alec Scott"
__license__ = "Apache-2.0"

import pakages.utils as utils
import pakages.build
from pakages.logger import logger
from citelang.main.packages import SetupManager

from cyclonedx_py.parser.requirements import RequirementsParser
from cyclonedx.model.bom import Bom
from cyclonedx.output import get_instance, OutputFormat
from cyclonedx.model import Tool

from pakages.version import __version__ as version
import shutil
import os


class RPackage:
    def __init__(self, path):
        self.path = path

    @classmethod
    def matches(cls, root):
        """
        Determine if a root has a setup.py or setup.cfg
        """
        files = os.listdir(root)
        return "setup.py" in files or "setup.cfg" in files

R -e "install.packages('remotes')"
R -e "remotes::install_local('/path/to/mypackage.tar.gz', dependencies=T)"

    def build(self):
        """
        A build returns a build result.
        """
        # The build result will manage the result files
        tmpdir = utils.get_tmpdir()
        result = pakages.build.BuildResult("python", tmpdir)
        bom = None

        with utils.workdir(self.path):

            # Build distribution
            command = f"python setup.py sdist --dist-dir {tmpdir}"
            cmd = utils.Command(command)
            res = cmd.run()

            # If we have a setup.py, generate an sbom
            if os.path.exists("setup.py"):
                manager = SetupManager()
                reqs = manager.from_file("setup.py")
                requirements = [
                    "%s==%s" % (x["name"], x["number"]) for x in reqs["dependencies"]
                ]
                rp = RequirementsParser("\n".join(requirements))
                bom = Bom.from_parser(parser=rp)

        logger.info(res.output)
        logger.warning(res.error)
        if res.returncode != 0:
            shutil.rmtree(tmpdir)
            logger.exit(f"Error running command: {command}.")
        targz = os.listdir(tmpdir)
        if not targz:
            logger.exit("No files found in build directory.")
        targz_path = os.path.join(tmpdir, targz[0])
        result.add_archive(targz_path, "application/vnd.oci.image.layer.v1.tar+gzip")

        if bom:
            bom.metadata.tools.add(
                Tool(vendor="Syspack", name="pakages", version=version)
            )
            instance = get_instance(bom, output_format=OutputFormat.JSON)
            sbom_path = os.path.join(tmpdir, "sbom.json")
            instance.output_to_file(sbom_path)
            result.add_archive(sbom_path, "application/vnd.cyclonedx")

        return result



__author__ = "Vanessa Sochat, Alec Scott"
__copyright__ = "Copyright 2021-2022, Vanessa Sochat and Alec Scott"
__license__ = "Apache-2.0"

from pakages.logger import logger
import pakages.oras
import shutil
import os


class BuildResult:
    """
    A BuildResult holds a set of content types and objects to upload.

    We also store the builder name (e.g., Python) so we know how to extract.
    """

    def __init__(self, builder, tmpdir, annotations=None):
        self.builder = builder
        self.archives = {}
        self.tmpdir = tmpdir
        self.titles = {}
        self.annotations = annotations or {}

    def __repr__(self):
        return str(self)

    def __str__(self):
        return "[pakages-build-task]"

    def remove(self):
        """
        Delete temporary directory
        """
        if self.tmpdir and os.path.exists(self.tmpdir):
            shutil.rmtree(self.tmpdir)

    def push(self, target):
        """
        Push archives to a target
        """
        reg = pakages.oras.get_oras_client()
        reg.push(target, self.archives, self.annotations, titles=self.titles)

    def summary(self):
        """
        Print a summary of files
        """
        logger.info(
            "Generated %s build artifacts for %s builder."
            % (len(self.archives), self.builder)
        )
        for path, mediaType in self.archives.items():
            logger.info("%s %s" % (mediaType.ljust(50), path))

    def add_title(self, path, title):
        """
        Add a specific title for a blob.
        """
        self.titles[path] = title

    def add_archive(self, path, mediaType):
        """
        Add an archive to the build result
        """
        self.archives[path] = mediaType



__author__ = "Vanessa Sochat, Alec Scott"
__copyright__ = "Copyright 2021-2022, Vanessa Sochat and Alec Scott"
__license__ = "Apache-2.0"

import itertools
import multiprocessing
import os
import time
import signal
import sys
from pakages.logger import logger
from typing import Optional


class Workers:
    def __init__(self, workers=None):

        if workers is None:
            workers = int(os.environ.get("PAKAGES_WORKERS", 9))
        self.workers = workers
        logger.debug(f"Using {self.workers} workers for multiprocess.")

    def start(self):
        logger.debug("Starting multiprocess")
        self.start_time = time.time()

    def end(self):
        self.end_time = time.time()
        self.runtime = self.runtime = self.end_time - self.start_time
        logger.debug(f"Ending multiprocess, runtime: {self.runtime} sec")

    def run(self, funcs: dict, tasks: dict) -> Optional[dict]:
        """
        Run will send a list of tasks, a tuple with arguments, through a function.
        the arguments should be ordered correctly.

        Args:
            - funcs (dict) : the functions to run with multiprocessing.pool,
                             a dictionary with lookup by the task name
            - tasks (dict) : a dict of tasks, each task name (key) with a tuple
                             of arguments to process
        """
        # Number of tasks must == number of functions
        assert len(funcs) == len(tasks)

        # Keep track of some progress for the user
        progress = 1

        # if we don't have tasks, don't run
        if not tasks:
            return None

        # results will also have the same key to look up
        finished = dict()
        results = []

        try:
            pool = multiprocessing.Pool(self.workers, init_worker)

            self.start()
            for key, params in tasks.items():
                func = funcs[key]
                result = pool.apply_async(multi_wrapper, multi_package(func, [params]))

                # Store the key with the result
                results.append((key, result))

            while len(results) > 0:
                pair = results.pop()
                key, result = pair
                result.wait()
                progress += 1
                finished[key] = result.get()

            self.end()
            pool.close()
            pool.join()

        except (KeyboardInterrupt, SystemExit):
            logger.error("Keyboard interrupt detected, terminating workers!")
            pool.terminate()
            sys.exit(1)

        except:
            logger.error("Error running task")

        return finished


# Supporting functions for MultiProcess Worker
def init_worker():
    signal.signal(signal.SIGINT, signal.SIG_IGN)


def multi_wrapper(func_args):
    function, kwargs = func_args
    return function(**kwargs)


def multi_package(func, kwargs):
    zipped = zip(itertools.repeat(func), kwargs)
    return zipped



__author__ = "Vanessa Sochat, Alec Scott"
__copyright__ = "Copyright 2021-2022, Vanessa Sochat and Alec Scott"
__license__ = "Apache-2.0"

import os
import oras.oci
import oras.defaults
import oras.provider
from oras.decorator import ensure_container
from pakages.logger import logger


def get_oras_client():
    """
    Consistent method to get an oras client
    """
    user = os.environ.get("ORAS_USER")
    password = os.environ.get("ORAS_PASS")
    reg = Registry()
    if user and password:
        print("Found username and password for basic auth")
        reg.set_basic_auth(user, password)
    else:
        logger.warning("ORAS_USER or ORAS_PASS is missing, push may have issues.")
    return reg


def pull_to_dir(pull_dir, target):
    """
    Given a URI, pull to an output directory.
    """
    reg = get_oras_client()
    return reg.pull(target=target, outdir=pull_dir)


class Registry(oras.provider.Registry):
    @ensure_container
    def push(self, container, archives: dict, annotations=None, titles: dict = None):
        """
        Given a dict of layers (paths and corresponding mediaType) push.
        """
        # Lookup of titles to override default
        titles = titles or {}

        # Prepare a new manifest
        manifest = oras.oci.NewManifest()

        # A lookup of annotations we can add
        annotset = oras.oci.Annotations(annotations or {})

        # Upload files as blobs
        for blob, mediaType in archives.items():

            # Must exist
            if not os.path.exists(blob):
                logger.exit(f"{blob} does not exist.")

            # Save directory or blob name before compressing
            blob_name = os.path.basename(blob)
            if blob in titles:
                blob_name = titles[blob]

            # If it's a directory, we need to compress
            cleanup_blob = False
            if os.path.isdir(blob):
                blob = oras.utils.make_targz(blob)
                cleanup_blob = True

            # Create a new layer from the blob
            layer = oras.oci.NewLayer(blob, mediaType, is_dir=cleanup_blob)
            annotations = annotset.get_annotations(blob)
            layer["annotations"] = {oras.defaults.annotation_title: blob_name}
            if annotations:
                layer["annotations"].update(annotations)

            # update the manifest with the new layer
            manifest["layers"].append(layer)

            # Upload the blob layer
            logger.info(f"Uploading {blob} to {container.uri}")
            response = self._upload_blob(blob, container, layer)
            self._check_200_response(response)

            # Do we need to cleanup a temporary targz?
            if cleanup_blob and os.path.exists(blob):
                os.remove(blob)

        # Add annotations to the manifest, if provided
        manifest_annots = annotset.get_annotations("$manifest")
        if manifest_annots:
            manifest["annotations"] = manifest_annots

        # Prepare the manifest config (temporary or one provided)
        config_annots = annotset.get_annotations("$config")
        conf, config_file = oras.oci.ManifestConfig()

        # Config annotations?
        if config_annots:
            conf["annotations"] = config_annots

        # Config is just another layer blob!
        response = self._upload_blob(config_file, container, conf)
        self._check_200_response(response)

        # Final upload of the manifest
        manifest["config"] = conf
        self._check_200_response(self._upload_manifest(manifest, container))
        print(f"Successfully pushed {container}")
        return response



__author__ = "Vanessa Sochat, Alec Scott"
__copyright__ = "Copyright 2021-2022, Vanessa Sochat and Alec Scott"
__license__ = "Apache-2.0"

import os
import pakages.builders
from .settings import Settings


def get_client(builder=None, settings_file=None):
    """
    Get a pakages client
    """
    # We don't require other builders to have spack installed
    if builder == "spack":
        from pakages.builders.spack import SpackClient

        return SpackClient(settings_file=settings_file)
    return PakagesClient(settings_file=settings_file)


class PakagesClient:
    """
    Pakages has a main controller for interacting with pakages.
    """

    def __init__(self, *args, **kwargs):
        settings_file = kwargs.get("settings_file")
        validate = kwargs.get("validate", True)
        if not hasattr(self, "settings"):
            self.settings = Settings(settings_file, validate=validate)

    def __repr__(self):
        return str(self)

    def __str__(self):
        return "[pakages-client]"

    def push(self, uri, cache_dir=None, tag=None):
        """
        Given an existing cache directory, push known specs to a specific uri
        """
        raise NotImplementedError

    def build(self, *args, **kwargs):
        """
        Build one or more packages.
        """
        args = list(args)
        if not args[0] or args[0] == ".":
            args[0] = os.getcwd()
        pkg = pakages.builders.get_package(args[0])

        # This returns a build result
        result = pkg.build()
        result.summary()
        return result

    def install(self, packages, registry=None, tag=None, **kwargs):
        """
        Install one or more packages.
        """
        raise NotImplementedError

    def uninstall(self, packages):
        """
        Uninstall a pakage
        """
        raise NotImplementedError



__author__ = "Vanessa Sochat, Alec Scott"
__copyright__ = "Copyright 2021-2022, Vanessa Sochat and Alec Scott"
__license__ = "Apache-2.0"

from pakages.logger import logger
import subprocess
import os
import shlex
import shutil


def which(software, strip_newline=True):
    """get_install will return the path to where Singularity (or another
    executable) is installed.
    """
    return shutil.which(software)


def check_install(software, quiet=True, command="--version"):
    """check_install will attempt to run the singularity command, and
    return True if installed. The command line utils will not run
    without this check.

    Parameters
    ==========
    software: the software to check if installed
    quiet: should we be quiet? (default True)
    command: the command to use to check (defaults to --version)
    """
    cmd = [software, command]
    try:
        version = run_command(cmd, software)
    except:  # FileNotFoundError
        return False
    if version:
        if not quiet and version["return_code"] == 0:
            version = version["message"]
            logger.info("Found %s version %s" % (software.upper(), version))
        return True
    return False


def get_installdir():
    """get_installdir returns the installation directory of the application"""
    return os.path.abspath(os.path.dirname(os.path.dirname(__file__)))


def get_userhome():
    """get the user home based on the effective uid. If import of pwd fails
    (not supported for Windows) then fall back to environment variable.
    """
    try:
        import pwd

        return pwd.getpwuid(os.getuid())[5]
    except:
        return os.environ.get("HOME") or os.environ.get("HOMEPATH")


def get_user():
    """Get the name of the user. We first try to import pwd, but fallback to
    extraction from the environment.
    """
    try:
        import pwd

        return pwd.getpwuid(os.getuid())[0]
    except:
        return os.environ.get("USER") or os.environ.get("USERNAME")


def stream_command(cmd):
    """stream a command (yield) back to the user, as each line is available.

    # Example usage:
    results = []
    for line in stream_command(cmd):
        print(line, end="")
        results.append(line)

    Parameters
    ==========
    cmd: the command to send, should be a list for subprocess
    """
    process = subprocess.Popen(
        cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE, universal_newlines=True
    )

    # Allow the runner to choose streaming output or error
    stream = process.stdout.readline

    # Stream lines back to the caller
    for line in iter(stream, ""):
        yield line

    # If there is an error, raise.
    process.stdout.close()
    return_code = process.wait()
    if return_code != 0:
        logger.error(process.stderr.read().strip("\n"))
        raise subprocess.CalledProcessError(return_code, cmd)


def run_command(cmd):
    """run_command uses subprocess to send a command to the terminal.

    Parameters
    ==========
    cmd: the command to send, should be a list for subprocess
    error_message: the error message to give to user if fails,
    if none specified, will alert that command failed.

    """
    if not isinstance(cmd, list):
        cmd = shlex.split(cmd)

    output = subprocess.Popen(cmd, stderr=subprocess.STDOUT, stdout=subprocess.PIPE)
    t = output.communicate()[0], output.returncode
    output = {"message": t[0], "return_code": t[1]}

    if isinstance(output["message"], bytes):
        output["message"] = output["message"].decode("utf-8")
    return output


def confirm_action(question, force=False):
    """confirm if the user wants to perform a certain action

    Parameters
    ==========
    question: the question that will be asked
    force: if the user wants to skip the prompt
    """
    if force is True:
        return True

    response = input(question)
    while len(response) < 1 or response[0].lower().strip() not in "ynyesno":
        response = input("Please answer yes or no: ")

    if response[0].lower().strip() in "no":
        return False

    return True



__author__ = "Vanessa Sochat, Alec Scott"
__copyright__ = "Copyright 2021-2022, Vanessa Sochat and Alec Scott"
__license__ = "Apache-2.0"

from pakages.logger import logger
import pakages.utils
import pakages.build
import pakages.defaults
import pakages.settings

import shutil
import os


class BuildCache:
    """
    A controller that makes it easy to create a build cache and install to it.
    We require that it is specific to one spec for now.
    """

    def __init__(
        self, spec_name, cache_dir=None, username=None, email=None, settings=None
    ):
        if not cache_dir:
            cache_dir = pakages.utils.get_tmpdir()
        self.cache_dir = cache_dir

        # Inherit settings from the client, or set to empty settings
        if not settings:
            settings = pakages.settings.EmptySettings()
        self.settings = settings

        # Set when we do create to remember specs
        self.spec_string = spec_name
        self.spec = None

    def remove(self):
        """
        Delete the entire build cache
        """
        if self.cache_dir and os.path.exists(self.cache_dir):
            logger.warning("Removing %s" % self.cache_dir)
            shutil.rmtree(self.cache_dir)

    def add_as_mirror(self, name):
        """
        Add the cache directory to spack as a filesystem mirror

        This currently downloads and adds to /tmp, and we issue a warning to the
        user. I'm not sure how people will want to install from caches but for
        the time being I'm making it a temporary interaction.
        """
        commands = [
            ["spack", "mirror", "add", name, self.cache_dir],
            ["spack", "buildcache", "update-index", "-d", self.cache_dir],
            ["spack", "buildcache", "list", "--allarch"],
        ]
        for command in commands:
            try:
                for line in pakages.utils.stream_command(command):
                    logger.info(line.strip("\n"))

            # Mirror already exists, just pass for now
            except:
                pass

    def create(self, specs, key=None):
        """
        Create the build cache with some number of specs

        Ideally we could do spack buildcache add but that isn't supported.
        """
        # Save the spec name for pushing to remote cache later
        self.spec_name = specs
        command = ["spack", "buildcache", "create", "-r", "-a", "-u"]
        if key:
            command += ["-k", key]
        command += ["-d", self.cache_dir, specs]
        logger.info(command)
        for line in pakages.utils.stream_command(command):
            logger.info(line.strip("\n"))

    def load_spec(self):
        """
        Given a name at creation, find the corresponding spec json to load
        """
        if not self.spec_string:
            return

        # This might have an issue with a compiler as package
        for blob in pakages.utils.recursive_find(self.cache_dir, pattern="spec.json"):
            print(f"Looking for {self.spec_name} in {blob}")
            if self.spec_name in blob:
                print("Found!")
                self.spec = pakages.utils.read_json(blob)
                break

    def get_spec_uri(self):
        """
        Given a spec, return the unique resource identifier for the remote cache.
        """
        if not self.spec:
            self.load_spec()
        spec = self.spec["spec"]["nodes"][0]
        name = spec["name"]
        version = spec["version"]
        compiler = "%s-%s" % (spec["compiler"]["name"], spec["compiler"]["version"])
        target = spec["arch"]["target"]
        platform = spec["arch"]["platform"]
        osys = spec["arch"]["platform_os"]
        # TODO should we remove target?
        return f"{platform}-{osys}-{compiler}-{target}-{name}-{version}"

    def push(self, uri=None, tag=None):
        """
        Push the build cache to an OCI registry (compatible with ORAS)
        """
        tag = tag or self.settings.default_tag
        uri = uri or self.settings.trusted_packages_registry

        # Ensure the cache_dir exists
        if not os.path.exists(self.cache_dir) or not os.listdir(self.cache_dir):
            logger.exit(f"{self.cache_dir} is empty, use build to populate first.")

        # We must have the spec to generate archive name
        self.load_spec()
        if not self.spec:
            logger.exit(f"Cannot find spec for {self.spec_name} in cache.")

        if not uri:
            uri = self.get_spec_uri()
        if ":" not in uri:
            uri = f"{uri}:{tag}"

        # Create a build result with the temporary directory
        result = pakages.build.BuildResult("spack", self.cache_dir)

        # Here we add each filename (relative to the root) as a layer
        for blob in pakages.utils.recursive_find(self.cache_dir):

            # A title for the blob instead of the default basename
            result.add_title(blob, blob.replace(self.cache_dir + os.sep, ""))
            result.add_archive(blob, "application/vnd.oci.image.layer.v1.tar")

        result.push(uri)
        return result

    def __repr__(self):
        return str(self)

    def __str__(self):
        return "[pakages-build-cache]"



__author__ = "Vanessa Sochat, Alec Scott"
__copyright__ = "Copyright 2021-2022, Vanessa Sochat and Alec Scott"
__license__ = "Apache-2.0"

from .fileio import read_file
from .terminal import get_user, run_command
from pakages.logger import logger

from datetime import datetime
import os
import locale
import tempfile
import subprocess
import shutil
import shlex
import uuid
import platform


class Capturing:
    """capture output from stdout and stderr into capture object.
    This is based off of github.com/vsoch/gridtest but modified
    to write files. The stderr and stdout are set to temporary files at
    the init of the capture, and then they are closed when we exit. This
    means expected usage looks like:

    with Capturing() as capture:
        process = subprocess.Popen(...)

    And then the output and error are retrieved from reading the files:
    and exposed as properties to the client:

        capture.out
        capture.err

    And cleanup means deleting these files, if they exist.
    """

    def __enter__(self):
        self.set_stdout()
        self.set_stderr()
        self.output = []
        self.error = []
        return self

    def set_stdout(self):
        self.stdout = open(tempfile.mkstemp()[1], "w")

    def set_stderr(self):
        self.stderr = open(tempfile.mkstemp()[1], "w")

    def __exit__(self, *args):
        self.stderr.close()
        self.stdout.close()

    @property
    def out(self):
        """
        Return output stream. Returns empty string if empty or doesn't exist.
        Returns (str) : output stream written to file
        """
        if os.path.exists(self.stdout.name):
            return read_file(self.stdout.name)
        return ""

    @property
    def err(self):
        """
        Return error stream. Returns empty string if empty or doesn't exist.
        Returns (str) : error stream written to file
        """
        if os.path.exists(self.stderr.name):
            return read_file(self.stderr.name)
        return ""

    def cleanup(self):
        for filename in [self.stdout.name, self.stderr.name]:
            if os.path.exists(filename):
                os.remove(filename)


class Command:
    def __init__(self, cmd=None):
        """
        set a unique id that includes executor name (type) and random uuid)
        """
        self.cmd = cmd
        self.taskid = str(uuid.uuid4())
        self.pwd = os.getcwd()
        self.message = None
        if not hasattr(self, "data"):
            self.data = {}

    def __str__(self):
        if self.cmd:
            return "[pakages.utils.Command:%s]" % self.cmd
        return "[pakages.utils.Command]"

    def _export_host(self):
        """
        Export host information from archspec.
        """
        host = {
            "architecture": platform.machine(),
            "python-compiler": platform.python_compiler(),
            "python-version": platform.python_version(),
            "cpu-count": platform.os.cpu_count(),
            "system": platform.os.uname().sysname,
        }
        libc = "@".join(platform.libc_ver())
        if libc != "@":
            host["libc"] = libc
        return {"host": host}

    def decode(self, line):
        """Given a line of output (error or regular) decode using the
        system default, if appropriate
        """
        loc = locale.getdefaultlocale()[1]

        try:
            line = line.decode(loc)
        except:
            pass
        return line

    def _export_common(self):
        """export common task variables such as present working directory, user,
        and timestamp for when it was run. This might include envars at some
        point, but we'd need to be careful.
        """
        common = {
            "pwd": self.pwd,
            "user": get_user(),
            "timestamp": str(datetime.now()),
        }
        if self.message:
            common["message"] = self.message

        # Try to get the version of the command
        cmd = self.data.get("cmd", [])
        if cmd:
            binary = shutil.which(cmd[0])
            if not binary:
                logger.exit(
                    "Error parsing command %s - did you forget the executable?"
                    % " ".join(cmd)
                )
            res = run_command([binary, "--version"])
            if res["return_code"] != 0:
                res = run_command([binary, "version"])
            if res["return_code"] == 0:
                common["version"] = res["message"].strip()

        return common

    def export(self):
        """
        Return data as json.
        """
        # Get common context (e.g., pwd)
        common = self._export_common()
        common.update(self._export_host())
        common.update(self.data)
        return common

    def get_command(self, cmd=None):
        cmd = cmd or self.cmd
        if not cmd:
            logger.exit("A command (cmd) is required.")
        if not isinstance(cmd, list):
            cmd = shlex.split(cmd)
        return cmd

    def run(self, cmd=None):
        """Run and capture a shell command.
        We use Capturing and then save attributes like the pid, output, error
        to it, and return to the calling function.
        """
        cmd = self.get_command(cmd)

        # Capturing provides temporary output and error files
        with Capturing() as capture:
            process = subprocess.Popen(
                cmd,
                stdout=capture.stdout,
                stderr=capture.stderr,
                universal_newlines=True,
            )
            capture.pid = process.pid
            returncode = process.poll()

            # Iterate through the output
            while returncode is None:
                returncode = process.poll()

        # Get the remainder of lines, add return code
        capture.output = self.decode(capture.out)
        capture.error = self.decode(capture.err)

        # Cleanup capture files and save final return code
        capture.cleanup()
        capture.returncode = returncode
        return capture

    def summary(self):
        return "[%s]" % self.name



__author__ = "Vanessa Sochat, Alec Scott"
__copyright__ = "Copyright 2021-2022, Vanessa Sochat and Alec Scott"
__license__ = "Apache-2.0"

from pakages.client import get_client
from .helpers import parse_extra


def main(args, parser, extra, subparser):
    kwargs = parse_extra(extra)
    cli = get_client(builder=args.builder, settings_file=args.settings_file)
    cli.install(args.packages, registry=args.registry, tag=args.tag, **kwargs)



from .python import PythonPackage
from pakages.logger import logger
import os


def get_package(root):
    """
    Get a packager based on a context
    """
    if isinstance(root, list):
        root = root[0]
    if os.path.exists(root):
        if PythonPackage.matches(root):
            return PythonPackage(root)
    logger.exit(f"We currently don't support a builder for content in {root}")



__author__ = "Vanessa Sochat, Alec Scott"
__copyright__ = "Copyright 2021-2022, Vanessa Sochat and Alec Scott"
__license__ = "Apache-2.0"

from pakages.logger import logger
import pakages.builders.spack.cache as spack_cache
import pakages.client
import pakages.oras

import spack.cmd
import spack.target
import spack.main
import spack.config

import re
import shlex
import os
import json


class SpackClient(pakages.client.PakagesClient):
    """
    Pakages has a main controller for interacting with pakages.
    """

    def parse_package_request(self, packages):
        """
        Parse the packages and repo (if any) from it.
        This is shared between install and build
        """
        # By defualt, assume not adding a repository
        repo = None

        if not isinstance(packages, list):
            packages = shlex.split(packages)

        # Case 1: we have an install directed at the present working directory
        if packages and packages[0] == ".":
            repo = os.getcwd()
            packages.pop(0)

        # If we have a path (akin to the first)
        if packages and os.path.exists(packages[0]):
            repo = packages.pop(0)

        # OR if we have a github URI TODO, can clone here
        if packages and re.search("(http|https)://github.com", packages[0]):
            repo = packages.pop(0)

        # If we don't have packages and we have a repo, derive from PWD
        if repo and not packages:
            packages = os.listdir(repo)

        # Finally, add the repository
        if repo:
            self.add_repository(repo)

        return packages

    def list_installed(self):
        """
        List installed packages
        """
        find = spack.main.SpackCommand("find")
        print(find())
        return json.loads(find("--json"))

    def build(self, packages, cache_dir=None, key=None, **kwargs):
        """
        Build a package into a cache
        """
        packages = self.parse_packages(packages)

        # Prepare a cache directory
        cache = spack_cache.BuildCache(
            spec_name=packages,
            cache_dir=cache_dir or self.settings.cache_dir,
            username=self.settings.username,
            email=self.settings.email,
            settings=self.settings,
        )

        # Install all packages
        self._install(packages)
        cache.create(packages, key=key)

        # Push function is on cache, if desired
        return cache

    def parse_packages(self, packages):
        """
        Helper function to ensure we return consistent names.
        """
        packages = self.parse_package_request(packages)
        if isinstance(packages, list):
            packages = packages[0]
        if " " in packages:
            logger.exit("We currently only support one package for build.")
        logger.info(f"Preparing package {packages}")
        return packages

    def add_repository(self, path):
        """
        Add a repository.

        Given a path that exists, add the repository to the
        underlying spack. If you need to add a GitHub uri, create a
        pakages.repo.PakRepo first.
        """
        repos = spack.config.get("repos")
        if path not in repos:
            logger.info(f"Adding repo {path} to spack repos.")
            repos.insert(0, path)
            spack.config.set("repos", repos)
            logger.info(f"Spack repos {repos}")

    def download_cache(self, target, download_dir=None):
        """
        Download a target to a cache download directory
        """
        download_dir = download_dir or pakages.utils.get_tmpdir()
        reg = pakages.oras.get_oras_client()

        # This will error if not successful, result is a list of files
        reg.pull(target=target, outdir=download_dir)
        return download_dir

    def install(self, packages, **kwargs):
        """
        Install one or more packages.
        """
        packages = self.parse_packages(packages)
        use_cache = kwargs.get("use_cache", False)
        if use_cache:
            cache_dir = self.download_cache(use_cache)
            cache = spack_cache.BuildCache(
                packages, cache_dir=cache_dir, settings=self.settings
            )

            # Cache is named after target, this is a filesystem mirror
            cache.add_as_mirror(re.sub("(-|:|/)", "-", use_cache))

        # Prepare install command with or without cache
        command = ["spack", "install"]
        if use_cache:
            command.append("--use-cache")
        command.append(" ".join(packages))

        # Install packages using system spack - we aren't responsible for this working
        for line in pakages.utils.stream_command(command):
            logger.info(line.strip("\n"))

    def _install(self, packages):
        """
        Install one or more packages.

        This eventually needs to take into account using the GitHub packages bulid cache
        """
        # Install packages using system spack - we aren't responsible for this working
        for line in pakages.utils.stream_command(["spack", "install", packages]):
            logger.info(line.strip("\n"))

    def uninstall(self, packages):
        """
        Uninstall a spack package
        """
        for line in pakages.utils.stream_command(["spack", "uninstall", packages]):
            logger.info(line.strip("\n"))
